#!/usr/bin/perl
#
# clusterDetails-inspect.pl - maps results of clustering by ELKI to 
#                             details of biflows in CSV
#                           - inspects clusters for different classes of
#                             attacks: SBA, MBA, DBA
#                           - outputs to the stdout
# 
# Copyright (C) 2012-2013, Masaryk University
# Author: Jan VYKOPAL <vykopal@ics.muni.cz>
#         
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 3. Neither the name of Masaryk University nor the names of its
#    contributors may be used to endorse or promote products derived 
#    from this software without specific prior written permission.
#
# This software is provided ``as is'', and any express or implied
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall Masaryk University or contributors be liable for
# any direct, indirect, incidental, special, exemplary, or consequential
# damages (including, but not limited to, procurement of substitute
# goods or services; loss of use, data, or profits; or business
# interruption) however caused and on any theory of liability, whether
# in contract, strict liability, or tort (including negligence or
# otherwise) arising in any way out of the use of this software, even
# if advised of the possibility of such damage.
#
# Version: 0.1
#
# $Id: clusterDetails-inspect.pl 1192 2013-04-18 20:36:50Z 98724 $

use Data::Dumper;

use strict;

my $threshold = 10;

my $ARGC = @ARGV;

if ($ARGC < 2) {
    die "usage: ./clusterDetails-inspect.pl <output_dir_ELKI> <input_file_details> <ratio>\n";
}

my $output_dir = $ARGV[0];

my $input_file_details = $ARGV[1];

my $ratio = $ARGV[2];

if ($ratio == '') {
	$ratio = 10;
	print "ratio was set to $ratio\n";
}
  
# with noise
# my @files = `ls $output_dir/cluster* $output_dir/noise.txt`;
# without noise
my @files = `ls $output_dir/cluster*`;

open DETAILS, "<$input_file_details" or die $!;
my @details = <DETAILS>;

foreach (@files) {
  open FILE, "<$_" or die $!;
  print "$_\n";
  my @clusters = <FILE>;

  my @sips;
  my @dips;

  my @pairs;

  my %sip_count = ();
  my %dip_count = ();

  my %dsts; # all destinations for a particular source, i. e. $dsts[$source] contains array of destinations for $source
  my %srcs; # all sources for a particular destination, i. e. $srcs[$destination] contains array of sources for $destination

  my %dsts_final;
  my %srcs_final;

  # get details from the CSV file
  foreach (@clusters) {
    if ($_ =~ m/ID=([0-9]+).*/) {
      my @biflow = split(/,/,$details[$1]);

      my $sip = $biflow[4];
      my $dip = $biflow[6];      
      
      push (@sips, $sip);
      push (@dips, $dip);

      ++$sip_count{$sip};
      ++$dip_count{$dip};

      push(@{$dsts{$sip}}, $dip);
      push(@{$srcs{$dip}}, $sip);
    }
  }

  # list pairs of IPs
  for (my $i=0; $i<scalar(@sips); $i++) {
   print $sips[$i], ' <-> ', $dips[$i], "\n";
  }
  print "\n";

  my $cluster_size = scalar (@sips);
  print "Cluster size: $cluster_size\n";

  # inspect %dsts  
  foreach my $src (sort keys %dsts) {
    my @destinations = @{$dsts{$src}};
    # print Dumper(\@destinations);

    my %counter;
    foreach my $dst (@destinations) {
      ++$counter{$dst};    
    }   

    foreach my $dst (sort keys %counter) {
	if ($counter{$dst} >= $threshold) {
	  $dsts_final{$src}{$dst} = $counter{$dst};
	}
    }    
  }

  # inspect %srcs
  foreach my $dst (sort keys %srcs) {
    my @sources = @{$srcs{$dst}};

    my %counter;
    foreach my $src (@sources) {
      ++$counter{$src};    
    }   

    foreach my $src (sort keys %counter) {
	if ($counter{$src} >= $threshold) {
	  $srcs_final{$dst}{$src} = $counter{$src};
	}
    }
  }

  print "\n";
  
  # classify attacks from %dsts_final using ratio - is there any MBA?
  foreach my $src (keys %dsts_final) {      
    my %destinations = %{$dsts_final{$src}};

    my @counts = (values %destinations);
    my $size = scalar @counts;

    if ($size >= $ratio) {    
      print "MBA from $src\nto: ";
      print Dumper(\%destinations);  
      }
  }

  # classify attacks from %srcs_final using ratio - is there any DBA?
  foreach my $dst (keys %srcs_final) {
    my %sources = %{$srcs_final{$dst}};

    my @ips = keys %sources;
    my @counts = values %sources;
    my $size = scalar @counts;

    if ($size >= $ratio) {     
      print "DBA to $dst\nfrom: ";
      print Dumper(\%sources);
    }
  }

  # classify attacks from %srcs_final - is there any SBA?
  foreach my $dst (keys %srcs_final) {
    my %sources = %{$srcs_final{$dst}};

    my @ips = keys %sources;
    
    my @counts = values %sources;
    my $size = scalar @counts;

    # number of hosts communicating with the IP
    my $size2 = scalar (keys %{$dsts_final{$ips[0]}});

    my $src_name = '';
    my $dst_name = '';

    # do the IPs communicate just to each other and aren't they part of MBA or SBA?
    if ($size == 1 && $size2 == 1 && $counts[0] == $dsts_final{$ips[0]}{$dst}) {
      print "SBA from ", $ips[0], " to $dst (", $counts[0], ")\n";
    }
  }
  print "\n";

  close FILE;
}

close DETAILS;

sub uniq {
  my %seen;
  return grep { !$seen{$_}++ } @_;
}
