#!/usr/bin/perl
#
# nfcapd2biflowscsv.pl - pairs unidirectional flows and produces biflows
#                      - the time complexity is quadratic!
#                      - outputs CSV with volume characteristics of the 
#                        biflows and CSV will all details of the biflows
#                      - input flows can be specified by a nfdump filter 
# 
# Copyright (C) 2012-2013, Masaryk University
# Author: Jan VYKOPAL <vykopal@ics.muni.cz>
#         
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 3. Neither the name of Masaryk University nor the names of its
#    contributors may be used to endorse or promote products derived 
#    from this software without specific prior written permission.
#
# This software is provided ``as is'', and any express or implied
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall Masaryk University or contributors be liable for
# any direct, indirect, incidental, special, exemplary, or consequential
# damages (including, but not limited to, procurement of substitute
# goods or services; loss of use, data, or profits; or business
# interruption) however caused and on any theory of liability, whether
# in contract, strict liability, or tort (including negligence or
# otherwise) arising in any way out of the use of this software, even
# if advised of the possibility of such damage.
#
# Version: 0.1
#
# $Id: nfcapd2biflowcsv.pl 1178 2013-04-17 06:58:06Z 98724 $

use strict;

use Date::Manip;
use Time::HiRes;

my $ARGC = @ARGV;

if ($ARGC != 4) {
    die "usage: ./nfcapd2biflowcsv.pl <input_file> <filter> ",
        "<time_window> <output_file>\n";
}

# Set time counter
my $total_time = Time::HiRes::time;
       
my $nfdump = '/usr/bin/nfdump';

my $input_file = $ARGV[0];
my $filter = $ARGV[1];
my $time_window = $ARGV[2]; # for instance: 2012/12/17.08:11-2012/12/17.08:12
my $output_file = $ARGV[3];
my $output_file_details = "details-$output_file";

my $nfdump_flows = "$nfdump -r $input_file -o 'fmt:%ts;%td;%pr;%sa;%sp;%da;%dp;%flg;%pkt;%byt' -q -m -N '$filter'";

# is time window set?
if ($time_window ne '') {
    $nfdump_flows .= " -t $time_window";
}

my @flows = `$nfdump_flows`;

my $time_counter = 0;
my $flow_count = 0;
my $biflow_count = 0;
my $time_diff = 1000; # maximal time difference between biflows in ms       
my $swap_counter = 0;

# save volume characteristics and all details of biflow into CSV
open FILE, ">$output_file" or die $!;
open FILE_DETAILS, ">$output_file_details" or die $!;

print FILE "pkt_fwd,pkt_rvs,byt_fwd,byt_rvs\n";
print FILE_DETAILS "id,ts,null,proto,sip,sp,dip,dp,null,pkt_fwd,pkt_rvs,byt_fwd,byt_rvs\n";

STDOUT->printflush("Pairing ", scalar(@flows), " flows (one dot per 2 % of processed flows) ");

for (my $i = 0; $i < @flows; $i++) {

    # extract flow features and statistics
    my @flow = split(/;/,$flows[$i]);
    my $ts = $flow[0];
    # TODO: extract flow duration
    my $proto = trim($flow[2]);
    my $sip = trim(substr $flow[3], 1);
    my $sp = trim(substr $flow[4], 1);
    my $dip = trim(substr $flow[5], 1);
    my $dp = trim(substr $flow[6], 1);
    my $fl = trim(substr $flow[7], 1);
    my $pkt = trim(substr $flow[8], 1);
    my $byt = trim(substr $flow[9], 1);

    for (my $j = $i+1; $j < @flows; $j++) {

		# extract flow features and statistics
        my @flow2 = split(/;/,$flows[$j]);
        my $ts2 = $flow2[0];
        # TODO: extract flow duration
        my $proto2 = trim($flow2[2]);
        my $sip2 = trim(substr $flow2[3], 1);
        my $sp2 = trim(substr $flow2[4], 1);
        my $dip2 = trim(substr $flow2[5], 1);
        my $dp2 = trim(substr $flow2[6], 1);                        
        my $fl2 = trim(substr $flow2[7], 1);
        my $pkt2 = trim(substr $flow2[8], 1);
        my $byt2 = trim(substr $flow2[9], 1);

        # is it a biflow?
        if ($sip eq $dip2 && $dip eq $sip2 && $proto == $proto2 && $sp == $dp2 && $dp == $sp2) {
			
			# time difference of flow start times
            my $diff = msepoch($ts2) - msepoch($ts);
            $time_counter++ if ($diff <= $time_diff);
            $biflow_count++;
            # print "biflow found: $ts - $ts2  = $diff, $proto $sip $sp $dip $dp\n";
            
            # is the forward and reverse direction of the found biflow swaped?
            # beware of protocols using a high service port by default, e. g., RDP using 3389
            if (($sp < 1024 && $dp > 1023) || ($sp == 3389 and $dp != 3389)) {
                # print "swap biflow? $sip:$sp<->$dip:$dp\n";
                my $tmp_sp = $sp;
                my $tmp_sip = $sip;
                $sp = $dp;
                $sip = $dip;
                $dp = $tmp_sp;
                $dip = $tmp_sip;
                $swap_counter++;
                # print "biflow swapped: $sip:$sp<->$dip:$dp\n";
            }

			# print volume characteristics of the biflow to the output file
            print FILE "$pkt,$pkt2,$byt,$byt2\n";
            # print all details of the biflow to the output file with details
            print FILE_DETAILS "$biflow_count,$ts,null,$proto,$sip,$sp,$dip,$dp,$fl|$fl2,$pkt,$pkt2,$byt,$byt2\n";
            last;
        }
         
    }
    
    STDOUT->printflush('.') if ($i % (scalar(@flows) / (100 / 2)) == 0);
}
print " DONE\n";
$flow_count = scalar (@flows);

print "\nStats:\n\nflows = $flow_count\nbiflows = $biflow_count\nbiflows with time_diff < $time_diff ms = $time_counter\nflows swaped = $swap_counter\n\n";

close FILE;
close FILE_DETAILS;

$total_time = sprintf ("%.6f", Time::HiRes::time - $total_time);
print "TOTAL RUNTIME $total_time sec.\n";
        
sub msepoch {
    my $unix_time = UnixDate(ParseDate($_[0]), "%s");
    my $ms = substr($_[0], -3, 3);
    return $unix_time * 1000 + $ms;
}

sub trim {
	my $string = $_[0];
        $string =~ s/^\s+//;
	$string =~ s/\s+$//;
        return $string;
}
